from datetime import timedelta
from unittest import TestCase
from uuid import uuid4

from proton import Message

from qpid_bow.config import configure
from qpid_bow.management.queue import create_queue
from qpid_bow.message import create_message
from qpid_bow.receiver import Receiver
from qpid_bow.sender import Sender

from . import TEST_AMQP_SERVER

CONFIG = {
    'amqp_url': f'amqp://127.0.0.1:5432, {TEST_AMQP_SERVER}'
}


class TestFailoverReceiver(TestCase):
    def setUp(self):
        configure(CONFIG)
        self.received_messages = []

        queue_address = uuid4().hex
        create_queue(queue_address, durable=False,
                     auto_delete=True, priorities=5,
                     extra_properties={'qpid.auto_delete_timeout': 10})
        self.sender = Sender(queue_address)

        def handle_received_message(message: Message):
            self.received_messages.append(message)
            return True
        self.receiver = Receiver(handle_received_message, queue_address)

    def test_failover_receive(self):
        expected_messages = (create_message(b'FOOBAR'),
                             create_message(b'FOOBAR'),
                             create_message(b'FOOBAR'))
        self.sender.queue(expected_messages)
        self.sender.send()

        self.receiver.limit = 3
        self.receiver.receive(timeout=timedelta(seconds=2))
        for received, expected in zip(self.received_messages,
                                      expected_messages):
            self.assertEqual(received.id, expected.id)
            self.assertEqual(received.body, expected.body)



import re

from qpid_bow.management.statistics import queue_statistics


def queue_stats_parser(action):
    parser = action.add_parser('stats', help="Shows queue statistics")
    parser.set_defaults(parser=parser)
    parser.set_defaults(func=queue_stats)

    parser.add_argument('-q', '--queue', type=str, required=False,
                        default=None, help="Queue address")
    parser.add_argument('-b', '--broker-url', type=str, required=False,
                        help="amqp:// or amqps:// URL to the broker")
    parser.add_argument('-f', '--filter', type=re.compile, required=False,
                        default=None, help="Filter by regex pattern")
    parser.add_argument('-a', '--autodelete',
                        action='store_true', default=False,
                        help="Include autodelete (temporary) queues")


def queue_stats(args):
    first = True
    line = "{name}\n\tMessage depth: {depth}\n\tTotal enqueued:{total}"

    try:
        queues = queue_statistics(args.queue, args.autodelete, args.broker_url)
    except ValueError as e:
        print(str(e) + '\n')
        args.parser.print_help()
        return

    for q_info in queues.values():
        if args.filter and not args.filter.match(q_info['name']):
            continue

        if first:
            first = False
        else:
            print('')  # newline
        print(line.format(**q_info))



from datetime import timedelta
from uuid import uuid4

from qpid_bow.exc import TimeoutReached
from qpid_bow.management.queue import create_queue
from qpid_bow.message import create_message
from qpid_bow.sender import Sender

from . import MessagingTestBase


class TestReceiver(MessagingTestBase):
    def test_receive_timeout(self):
        with self.assertRaises(TimeoutReached):
            self.receiver.receive(timeout=timedelta(seconds=2))
        self.assertEqual(len(self.received_messages), 0)

    def test_add_address_duplicate(self):
        second_queue_address = uuid4().hex
        self.receiver.add_address(second_queue_address)
        self.receiver.add_address(second_queue_address)
        self.assertEqual(len(self.receiver.receivers), 2)

    def test_multi_receive(self):
        expected_messages = [create_message(b'FOOBAR1'),
                             create_message(b'FOOBAR2'),
                             create_message(b'FOOBAR3')]
        self.send_messages(expected_messages[0:-1])

        second_queue_address = uuid4().hex
        create_queue(second_queue_address, durable=False, auto_delete=True,
                     extra_properties={'qpid.auto_delete_timeout': 10})
        self.receiver.add_address(second_queue_address)

        second_sender = Sender(second_queue_address)
        second_sender.queue(expected_messages[-1:])
        second_sender.send()

        self.receive_messages()
        self.assertEqual(len(self.received_messages), len(expected_messages))

    def test_receive(self):
        self.send_messages((create_message(b'FOOBAR1'),
                            create_message(b'FOOBAR2'),
                            create_message(b'FOOBAR3')))
        self.receive_messages()
        self.check_messages()

    def test_receive_limit(self):
        self.send_messages((create_message(b'FOOBAR1'),
                            create_message(b'FOOBAR2'),
                            create_message(b'FOOBAR3')))

        # Should first receive only 2
        self.receiver.limit = 2
        self.receive_messages()
        self.check_messages(expected_messages=self.expected_messages[:-1])

        # Receive the rest
        self.receiver.limit = None
        self.receive_messages()
        self.check_messages()



"""Receive messages from AMQP broker."""

import asyncio
import logging
from datetime import datetime, timedelta
from inspect import signature
from typing import Any, Awaitable, Callable, Optional, Type, Union
from uuid import uuid4

from proton import Delivery, Message
from proton.reactor import (
    Container,
    EventBase,
    Task,
)

from qpid_bow import Connector, ReconnectStrategy, RunState
from qpid_bow.exc import (
    QMF2Exception,
    RetriableMessage,
    TimeoutReached,
)

logger = logging.getLogger()


ReceiveCallback = Union[
    Callable[[Message], bool],
    Callable[[Message, Delivery], bool],
    Callable[[Message], Awaitable[bool]],
    Callable[[Message, Delivery], Awaitable[bool]],
]


class Receiver(Connector):
    """Callback based AMQP message receiver.

    Args:
        callback: Function to call when new message is received.
        address: Name of queue or exchange from where to receive the messages.
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.
        limit: Limit the amount of messages to receive.
        container_class: Qpid Proton reactor container-class to use.
        reconnect_strategy: Strategy to use on connection drop.
    """
    def __init__(
            self, callback: ReceiveCallback,
            address: Optional[str] = None,
            server_url: Optional[str] = None,
            limit: Optional[int] = None,
            container_class: Type[Any] = Container,
            reconnect_strategy: ReconnectStrategy = ReconnectStrategy.backoff
    ) -> None:
        super().__init__(server_url=server_url,
                         container_class=container_class,
                         reconnect_strategy=reconnect_strategy)
        self.limit = limit
        self.received = 0
        self.timeout: Optional[timedelta] = None
        self.timeout_task: Task = None
        self.receivers: dict = {}
        self.connection = None

        self.callback = callback
        self.advanced_callback = len(signature(callback).parameters) == 2
        self.timeout_reached = False
        self.start_time: Optional[datetime]

        if address:
            self.receivers[address] = None

    def receive(self, timeout: Optional[timedelta] = None):
        """Start receive loop for up to timeout duration or limit messages.

        Args:
            timeout: Timeout duration to wait for message.
        """
        self.timeout_reached = False
        self.timeout = timeout

        # Give control to container to do our receiving
        self.run()

        if self.timeout_reached:
            raise TimeoutReached()

    def add_address(self, address: str):
        """Start receiving messages from the given additional address.

        Args:
            address: Queue or exchange address to receive from.
        """
        if address in self.receivers:
            return

        if self.connection:
            self._start_receiver(address)
        else:
            self.receivers[address] = None

    def remove_address(self, address: str):
        """Stop receiving messages from the given address.

        Args:
            address: Queue or exchange address to stop receiving from.
        """
        receiver = self.receivers.pop(address)
        receiver.close()

    def _start_receiver(self, address: str):
        if address == '#':  # AMQP-dynamic queue address
            receiver = self.container.create_receiver(self.connection,
                                                      dynamic=True)
        else:
            receiver = self.container.create_receiver(
                self.connection,
                address,
                # Add UUID to name to prevent add/remove link race condition
                name=f'{self.connection.container}-{address}-{uuid4()}')
        self.receivers[address] = receiver

    def _restart_receivers(self):
        addresses = list(self.receivers.keys())
        self.receivers.clear()
        logger.debug("Starting receivers for addresses %s", addresses)
        for address in addresses:
            self._start_receiver(address)

    def stop(self):
        if self.timeout_task:
            self.timeout_task.cancel()
            self.timeout_task = None

        self.received = 0
        self.start_time = None
        for address in self.receivers:
            self.receivers[address].close()
            self.receivers[address] = None

        super().stop()

    def on_connection_opened(self, event: EventBase):
        previous_state = self.run_state
        super().on_connection_opened(event)
        if previous_state == RunState.reconnecting:
            logger.debug("AMQP transport reinstated, restarting receivers...")
            self._restart_receivers()

    def on_timer_task(self, event: EventBase):
        """Handles the event when a timer is finished.

        Args:
            event: Reactor timer task event object.
        """
        if self.run_state not in (RunState.started,
                                  RunState.connected,
                                  RunState.reconnecting):
            return

        if self.timeout is None or self.start_time is None:
            self.timeout_reached = True
            self.stop()
        elif (datetime.utcnow() - self.start_time) > self.timeout:
            self.timeout_reached = True
            self.stop()
        else:
            self.timeout_task = event.container.schedule(0.25, self)

    def on_start(self, event):
        super().on_start(event)
        if self.run_state == RunState.started:
            self.start_time = datetime.utcnow()
            if self.timeout:
                self.timeout_task = event.container.schedule(0.25, self)
            self._restart_receivers()

    def on_message(self, event):
        if not self.start_time:
            # Eagerly got more messages then we're interested in, release
            self.release(event.delivery)
            return

        try:
            if asyncio.iscoroutinefunction(self.callback):
                loop = asyncio.get_event_loop()
                if not loop.is_running():
                    success = loop.run_until_complete(
                        self.handle_async_message(event))
                else:
                    success = asyncio.ensure_future(
                        self.handle_async_message(event), loop=loop)
            else:
                success = self.handle_message(event)

            if success:
                self.accept(event.delivery)
            else:
                self.reject(event.delivery)
        except RetriableMessage:
            if event.message.delivery_count:
                self.reject(event.delivery)
            else:
                # Set the delivery status before releasing back into the queue
                # https://bugzilla.redhat.com/show_bug.cgi?id=1283652
                event.delivery.local.undeliverable = True
                event.delivery.local.failed = True
                self.release(event.delivery)
        except QMF2Exception:
            self.accept(event.delivery)
            raise
        except Exception:
            # On unexpected error, we will reject the message and re-raise
            logger.error('Unexpected error, rejecting the message',
                         exc_info=True)
            self.reject(event.delivery)
            raise
        finally:
            self.touch()
            self.received += 1
            if self.received == self.limit:
                self.stop()

    def handle_message(self, event):
        if self.advanced_callback:
            return self.callback(event.message, event.delivery)

        return self.callback(event.message)

    async def handle_async_message(self, event):
        if self.advanced_callback:
            return await self.callback(event.message, event.delivery)

        return await self.callback(event.message)



"""Qpid-bow client framework."""

from asyncio import Event as AsyncioEvent
from enum import Enum, auto
from logging import getLogger
from typing import Optional, Type

from proton import Connection
from proton.handlers import MessagingHandler
from proton.reactor import Container, Backoff, EventBase

from qpid_bow.config import (
    config,
    get_urls,
)

logger = getLogger()


class Priority(Enum):
    """Convenience enum for message priorities.

    Qpid supports a configurable amount of priorities for a queue,
    be sure to have at least 5.

    When used on a message and enabled on a queue Qpid will re-order which
    get send out to a receiver first based on the priority.
    """
    internal_low = 0
    low = 1
    normal = 2
    high = 3
    realtime = 4


class RunState(Enum):
    """Indicate current state of Connector."""
    stopped = auto()
    stopping = auto()
    started = auto()
    reconnecting = auto()
    connected = auto()
    failed = auto()


class NonBackoff(Backoff):
    def next(self):
        return self.delay


class ReconnectStrategy(Enum):
    """Define possible reconnect strategies."""
    backoff = Backoff()
    failover = NonBackoff()
    disabled = False


class Connector(MessagingHandler):
    """Initiate and keep connection to AMQP message broker.

    Args:
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.
        container_class: Qpid Proton reactor container-class to use.
        reconnect_strategy: Strategy to use on connection drop.
    """
    def __init__(
            self, server_url: Optional[str] = None,
            container_class: Type[Container] = Container,
            reconnect_strategy: ReconnectStrategy = ReconnectStrategy.backoff
        ) -> None:
        super().__init__(auto_accept=False)
        self.server_urls = get_urls(server_url)

        self.run_state = RunState.stopped
        self.failover_count = 0
        self.container_class = container_class
        self.reconnect_strategy = reconnect_strategy
        self.close_event = AsyncioEvent()
        self.close_event.set()
        self.connection: Connection
        self.container: Container

    def touch(self):
        """Instruct the reactor container to do processing.

        When running with an alternative container, like the AsyncioContainer,
        you might need to call this to startup new sessions.
        """
        try:
            self.container.touch()
        except (AttributeError, NameError):
            pass

    def on_start(self, event: EventBase):
        """Handle start event.

        Args:
            event: Reactor init event object with container to connect to.
        """
        super().on_start(event)
        if self.run_state == RunState.stopped:
            logger.debug("Starting %s", self)
            self.run_state = RunState.started
            self.connection = event.container.connect(
                urls=self.server_urls, reconnect=self.reconnect_strategy.value)

    def on_connection_opened(self, event: EventBase):  # pylint: disable=unused-argument
        self.run_state = RunState.connected
        self.close_event.clear()

    def on_transport_error(self, event: EventBase):
        super().on_transport_error(event)
        if self.reconnect_strategy == ReconnectStrategy.backoff:
            logger.warning("AMQP transport was closed, reconnecting...")
            self.run_state = RunState.reconnecting
            return
        elif self.reconnect_strategy == ReconnectStrategy.failover:
            self.failover_count += 1
            if (len(self.server_urls) * 2) > self.failover_count:
                return

        self.run_state = RunState.failed
        condition = event.transport.condition
        raise ConnectionError(f"{condition.name} {condition.description}")

    def on_connection_closed(self, event: EventBase):
        """Handle close connection event.

        Args:
            event: Connection close event.
        """
        logger.debug("Connection %s confirmed closed", self)
        self.run_state = RunState.stopped
        self.close_event.set()

    async def wait_closed(self):
        await self.close_event.wait()

    def run(self):
        """Start this Connector and setup connection to the AMQP server."""
        self.container = self.container_class(self)
        self.container.run()

    def stop(self):
        """Stop connection to the AMQP server."""
        if self.run_state not in (RunState.started,
                                  RunState.connected,
                                  RunState.reconnecting):
            return

        logger.debug("Connection %s closing", self)
        self.run_state = RunState.stopping
        self.connection.close()
        self.connection = None

        # Give control to container to cleanup
        try:
            self.container.touch()
        except (AttributeError, NameError):
            self.container.process()



from os import EX_USAGE

from qpid_bow.management.queue import reroute_queue


def queue_reroute_parser(action):
    parser = action.add_parser(
        'reroute', help="Reroute messages from a queue to an exchange, "
                        "optionally filtering")
    parser.set_defaults(parser=parser)
    parser.set_defaults(func=queue_reroute)

    parser.add_argument('queue', metavar='QUEUE', type=str,
                        help="Queue address")
    parser.add_argument('-b', '--broker-url', type=str, required=False,
                        help="amqp:// or amqps:// URL to the broker")
    parser.add_argument('-e', '--exchange-address', type=str, required=False,
                        help="Exchange address, if not provided"
                             "alternative exchange will be used")
    parser.add_argument('-l', '--limit', type=int, required=False, default=0,
                        help="Limit the number of messages")
    parser.add_argument('-f', '--filter', type=str, default=None, nargs='?',
                        help="Filter based on property=value")


def queue_reroute(args):
    try:
        filter_key, filter_value = args.filter.split('=', 1)
        filter_tuple = (filter_key, filter_value)
    except ValueError:
        print("Invalid filter specified\n")
        args.parser.print_help()
        exit(EX_USAGE)
    except AttributeError:
        filter_tuple = None

    reroute_queue(args.queue_address, args.exchange_address, args.limit,
                  filter_tuple, args.broker_url)



"""Gather session-related data from AMQP broker."""

from collections import defaultdict
from datetime import timedelta
from typing import (
    MutableMapping,
    Optional,
)

from proton import Message

from qpid_bow.management import create_QMF2_query
from qpid_bow.remote_procedure import RemoteProcedure


def get_sessions(server_url: Optional[str] = None) -> dict:
    """Retrieve sessions from AMQP broker.

    Args:
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.

    Returns:
        dict: A dict mapping between session id and it's address.

    Example:
        >>> get_sessions()
        {'org.apache.qpid.broker:session:0x7fb8bc021ab0': {'address': '10.0.0.2:34814'}}
    """
    sessions = {}

    def _update_sessions(message: Message):
        for item in message.body:
            values = item['_values']
            connectionRef = values['connectionRef']['_object_name'].decode()
            sessions[item['_object_id']['_object_name'].decode()] = {
                'address': connectionRef.rsplit('-', 1)[1]}
        return True

    rpc = RemoteProcedure(_update_sessions, 'qmf.default.direct',
                          server_url)
    binding_query_message = create_QMF2_query('org.apache.qpid.broker',
                                              'session')
    rpc.call(binding_query_message, timedelta(seconds=5))
    return sessions


def get_outgoing_sessions_by_address(
        server_url: Optional[str] = None) -> MutableMapping[str, list]:
    """Retrieve outgoing sessions from AMQP broker.

    Args:
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.

    Returns:
        defaultdict: A dict mapping between address name and list of sessions.

    Example:
        >>> get_outgoing_sessions_by_address()
        {'8152f68b-c74a-4d22-8630-a89cf194d067_8152f68b-\
c74a-4d22-8630-a89cf194d067-2d808664-fe81-4da4-8258-288a7ff531ac': [{\
'session_id': 'org.apache.qpid.broker:session:0x7fb8bc021ab0','transfers': \
ulong(0)}]}
    """
    client_subscriptions: MutableMapping[str, list] = defaultdict(list)

    def _update_results(message: Message):
        for value in (item['_values'] for item in message.body):
            client_subscriptions[value['source'].decode()].append({
                'session_id': value['sessionRef']['_object_name'].decode(),
                'transfers': value['transfers']
            })
        return True

    rpc = RemoteProcedure(_update_results, 'qmf.default.direct',
                          server_url)
    binding_query_message = create_QMF2_query('org.apache.qpid.broker',
                                              'outgoing')
    rpc.call(binding_query_message, timedelta(seconds=5))
    return dict(client_subscriptions)



from datetime import timedelta
from typing import Any, Mapping, Optional

from proton import Message

from qpid_bow.exc import (
    ObjectNotFound,
    QMF2Exception,
)
from qpid_bow.remote_procedure import RemoteProcedure

QUEUE_ID_PREFIX = 'org.apache.qpid.broker:queue:'
EXCHANGE_ID_PREFIX = 'org.apache.qpid.broker:exchange:'


def create_QMF2_message() -> Message:
    """Factory function to create a base message for a QMF2 RPC call.

    Returns:
        Message: A base message for a QMF2 RPC call.
    """
    return Message(subject='broker', properties={
        'routing-key': 'broker',
        'x-amqp-0-10.app-id': 'qmf2',
        'method': 'request'
    })


def create_QMF2_query(package_name: str, class_name: str) -> Message:
    """Factory function to create a QMF2 object query.

    Args:
        package_name: Qpid internal package name to query.
        class_name: Qpid internal class name to query.

    Returns:
        Message: A message fully setup with a QMF2 object query RPC call.
    """
    message = create_QMF2_message()
    message.correlation_id = class_name
    message.properties['qmf.opcode'] = '_query_request'
    message.body = {
        '_what': 'OBJECT',
        '_schema_id': {
            '_package_name': package_name,
            '_class_name': class_name
        }
    }

    return message


def create_QMF2_method_invoke(object_id: dict,
                              method_name: str,
                              arguments: Mapping[str, Any]) -> Message:
    """Factory function to create a QMF2 object method call.

    Args:
        object_id: Qpid internal object ID.
        method_name: Name of the method to call.
        arguments: Mapping with key/value pairs of arguments for the method.

    Returns:
        Message: A message fully setup with a QMF2 method invoke RPC call.
    """
    message = create_QMF2_message()
    message.corrrelation_id = 'method'
    message.properties['qmf.opcode'] = '_method_request'
    message.body = {
        '_object_id': object_id,
        '_method_name': method_name,
        '_arguments': arguments
    }

    return message


def handle_QMF2_exception(message: Message):
    """Deserialises and raises a QMF2 exception from a reply, in case the
    QMF2 RPC call failed.

    Args:
        message: The QMF2-RPC reply message.
    """
    if message.properties.get('qmf.opcode') != '_exception':
        return True

    raise QMF2Exception.from_data(message.body['_values'])


def get_object(package_name: str, class_name: str, object_name: str,
               server_url: Optional[str] = None) -> dict:
    """Find a raw QMF2 object by type and name.

    Args:
        package_name: Qpid internal package name to query.
        class_name: Qpid internal class name to query.
        object_name: Name of the Qpid object to find.
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.

    Returns:
        dict: Raw QMF2 object.
    """

    object_: dict = {}

    def handle_response(message: Message):
        for result in message.body:
            if result['_values']['name'].decode() == object_name:
                object_.update(result)
                return True

        return True

    rpc = RemoteProcedure(handle_response,
                          'qmf.default.direct', server_url)
    rpc.call(create_QMF2_query(package_name, class_name),
             timedelta(seconds=5))

    if not object_:
        raise ObjectNotFound(class_name, object_name)

    return object_


def get_broker_id(server_url: Optional[str] = None) -> dict:
    """Get the full internal broker ID object.

    Args:
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.

    Returns:
        dict: Full internal broker ID object.
    """
    broker_id: dict = {}

    def handle_response(message: Message):
        broker_id.update(message.body[0]['_object_id'])
        return True

    rpc = RemoteProcedure(handle_response, 'qmf.default.direct', server_url)
    broker_query_message = create_QMF2_query('org.apache.qpid.broker',
                                             'broker')
    rpc.call(broker_query_message, timedelta(seconds=5))
    return broker_id



"""Remote procedure call handling."""

from datetime import timedelta
from logging import getLogger
from typing import Optional
from warnings import warn

from proton import Message

from qpid_bow import ReconnectStrategy, RunState
from qpid_bow.receiver import (
    Receiver,
    ReceiveCallback,
)

logger = getLogger()


class RemoteProcedure(Receiver):
    """This class can be used to handle a simple RPC pattern,
       sending a call message and waiting for a reply on a temporary queue
       and response handling through callbacks.

    Args:
        callback: Function to call when new message is received.
        address: Address of queue or exchange to send the messages to.
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.
        reconnect_strategy: Strategy to use on connection drop.
    """
    def __init__(
            self, callback: ReceiveCallback,
            address: str, server_url: Optional[str] = None,
            reconnect_strategy: ReconnectStrategy = ReconnectStrategy.failover
    ) -> None:
        super().__init__(callback, '#', server_url,
                         reconnect_strategy=reconnect_strategy)
        if self.reconnect_strategy == ReconnectStrategy.backoff:
            warn("Using ReconnectStrategy.backoff may cause Sender to block")
        self.send_address = address
        self.message: Message

    def call(self, message: Message, timeout: Optional[timedelta] = None):
        """Send RPC message and wait for call reply.
        Args:
            message: Message to send to RPC-address.
            timeout: Optional maximum timeout to wait for a reply.
        """
        self.message = message

        # Receiver.receive() triggers container run with timeout
        self.receive(timeout)

    def on_start(self, event):
        super().on_start(event)
        if self.run_state == RunState.started:
            event.container.create_sender(self.connection, self.send_address)

    def on_sendable(self, event):
        if not self.reply_to:
            # Wait until receiver has been set up
            return

        self.message.reply_to = self.reply_to
        event.sender.send(self.message)

        # We only need to send one message
        event.sender.close()

    def on_message(self, event):
        super().on_message(event)
        if 'partial' not in event.message.properties:
            self.stop()


    @property
    def reply_to(self):
        """Reply to address of our temporary queue."""
        return self.receivers['#'].remote_source.address



"""Send messages to AMPQ broker."""

import logging
from typing import Iterable, Optional
from uuid import uuid4
from warnings import warn

from proton import Message

from qpid_bow import Connector, ReconnectStrategy, RunState
from qpid_bow.exc import UnroutableMessage

logger = logging.getLogger()


class Sender(Connector):
    """Class to send messages in a batch to an AMQP address.

    Args:
        address: Address of queue or exchange to send the messages to.
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.
        reconnect_strategy: Strategy to use on connection drop.
    """

    def __init__(
            self, address: Optional[str] = None,
            server_url: Optional[str] = None,
            reconnect_strategy: ReconnectStrategy = ReconnectStrategy.failover
    ) -> None:
        super().__init__(server_url, reconnect_strategy=reconnect_strategy)
        if self.reconnect_strategy == ReconnectStrategy.backoff:
            warn("Using ReconnectStrategy.backoff may cause Sender to block")
        self.address = address
        self.send_queue: list = []

    def queue(self, messages: Iterable[Message]):
        """Enqueue messages that will be send on calling :obj:`send`."""
        if not self.address:
            if any((new_message.address is None for new_message in messages)):
                raise UnroutableMessage(
                    "A Sender with no address requires Message.address is set")

        self.send_queue.extend(messages)

    def send(self):
        """Send queued messages."""
        # Give control to container to do our sending
        if self.send_queue:
            self.run()

    def on_start(self, event):
        super().on_start(event)
        if self.run_state == RunState.started:
            event.container.create_sender(self.connection, self.address)

    def on_sendable(self, event):
        """Handles sendable event, sends all the messages in the send_queue."""
        if not self.connection:
            return

        message = self.send_queue.pop(0)
        message.id = uuid4()
        # TODO SHA
        event.sender.send(message)

        if not self.send_queue:
            # We are done sending, clear & return control
            self.stop()



from random import shuffle
from time import sleep
from uuid import uuid4

import pytest

from qpid_bow import Priority, ReconnectStrategy
from qpid_bow.exc import UnroutableMessage
from qpid_bow.message import create_message
from qpid_bow.sender import Sender

from . import MessagingTestBase


class TestSender(MessagingTestBase):
    def test_reconnect_strategy_backoff_warning(self):
        with pytest.warns(UserWarning):
            Sender(uuid4().hex, reconnect_strategy=ReconnectStrategy.backoff)

    def test_connection_error(self):
        sender = Sender(
            uuid4().hex, server_url='amqp://invalid:invalid@example')
        sender.queue((create_message(b'FOOBAR'),))
        with self.assertRaises(ConnectionError):
            sender.send()

    def test_send_properties(self):
        message = create_message(b'FOOBAR', {'foo': 'bar', 'baz': 123})
        self.send_messages((message,))
        self.receive_messages()
        self.check_messages()

    def test_send_priorities(self):
        messages_to_send = []
        for i in range(0, 30):
            messages_to_send.append(
                create_message(b'LOW', priority=Priority.low))
            messages_to_send.append(
                create_message(b'NORMAL', priority=Priority.normal))
            messages_to_send.append(
                create_message(b'HIGH', priority=Priority.high))

        # Send messages with priority in random order
        shuffle(messages_to_send)
        self.send_messages(messages_to_send)

        # Should be received in order
        self.expected_messages = sorted(messages_to_send, reverse=True,
                                        key=lambda message: message.priority)

        # Give Qpid a bit of time for ordering if needed
        sleep(3)
        self.receive_messages()

        # Don't check identity, because only the priority order is preserved
        self.check_messages(check_identity=False)

    def test_addressless_routable(self):
        message = create_message(b'FOOBAR')
        message.address = self.sender.address
        self.sender.address = None
        self.send_messages((message,))
        self.receive_messages()
        self.check_messages()

    def test_addressless_unroutable(self):
        with self.assertRaises(UnroutableMessage):
            sender = Sender()
            message = create_message(b'FOOBAR')
            sender.queue((message,))



import asyncio
import time
from logging import getLogger

from typing import List, Union

from proton import Connection, Receiver, Sender, Session, Url
from proton.reactor import Container as BaseContainer, LinkOption
from proton.handlers import IOHandler

logger = getLogger()


class AsyncioReactorHandler:
    """Qpid Proton Reactor Global Loop Handler for Python asyncio.

    This implementation will setup Qpid Proton's Selectables to use asyncio's
    writable/readable event handling.

    Based on Tornado implementation:
    https://qpid.apache.org/releases/qpid-proton-0.18.1/proton/python/examples/proton_tornado.py.html

    Args:
        loop: An asyncio event loop
        handler_base: An IO Handler
    """

    def __init__(self, loop=None, handler_base=None):
        self.loop = loop or asyncio.get_event_loop()
        self.io = handler_base or IOHandler()
        self._count = 0
        self._reactor = None

    def on_reactor_init(self, event):
        logger.debug("Reactor initted")
        self._reactor = event.reactor

    def on_reactor_quiesced(self, event):  # pylint: disable=no-self-use
        # check if we are still quiesced, other handlers of
        # on_reactor_quiesced could have produced events to process
        if not event.reactor.quiesced:
            return

        logger.debug("Reactor quiesced")
        event.reactor.yield_()

    def on_unhandled(self, name, event):  # pylint: disable=unused-argument
        event.dispatch(self.io)

    def _schedule(self, selectable):
        if selectable.deadline:
            logger.debug("Setting up schedule for %s", selectable)
            self.loop.call_later(
                selectable.deadline - time.time(),
                lambda: self._scheduled_selectable_expired(selectable))

    def _scheduled_selectable_expired(self, selectable):
        logger.debug("Scheduled selectable %s expired", selectable)
        selectable.expired()
        self._process()

    def _process(self):
        self._reactor.process()
        if not self._reactor.quiesced:
            self.loop.call_soon_threadsafe(self._process)

    def _selectable_readable(self, selectable):
        logger.debug("Readable callback: %s", selectable)
        selectable.readable()
        self._process()

    def _selectable_writable(self, selectable):
        logger.debug("Writable callback: %s", selectable)
        selectable.writable()
        self._process()

    def _setup_selectable(self, selectable):
        if selectable.reading:
            logger.debug("Setting up reader for %s", selectable)
            self.loop.add_reader(
                selectable.fileno(), self._selectable_readable, selectable)
        if selectable.writing:
            logger.debug("Setting up writer for %s", selectable)
            self.loop.add_writer(
                selectable.fileno(), self._selectable_writable, selectable)

    def _teardown_selectable(self, selectable):
        logger.debug("Resetting %s", selectable)
        self.loop.remove_writer(selectable.fileno())
        self.loop.remove_reader(selectable.fileno())

    def on_selectable_init(self, event):
        selectable = event.context
        if selectable.fileno() >= 0:
            self._setup_selectable(selectable)

        self._schedule(selectable)
        self._count += 1

    def on_selectable_updated(self, event):
        selectable = event.context
        if selectable.fileno() > 0:
            self._teardown_selectable(selectable)
            self._setup_selectable(selectable)

        self._schedule(selectable)

    def on_selectable_final(self, event):
        selectable = event.context
        if selectable.fileno() > 0:
            self._teardown_selectable(selectable)

        logger.debug("Selectable final %s", selectable)
        selectable.release()
        self._count -= 1
        if self._count == 0:
            self.loop.call_soon_threadsafe(self._stop)

    def _stop(self):
        if self._reactor:
            logger.debug("Stopping reactor")
            self._reactor.stop()


# Needs to be called Container for magic reasons :(
class Container(BaseContainer):
    """Asyncio event loop based Qpid Reactor container.

    Args:
        *handlers: One or more connectors

    Keyword Args:
        handler_base: An IO Handler.
        impl: Reactor implementation, default is pn_reactor.
    """
    def __init__(self, *handlers, **kwargs):
        loop = kwargs.get('loop', asyncio.get_event_loop())
        kwargs['global_handler'] = AsyncioReactorHandler(
            loop, kwargs.get('handler_base', None))
        super().__init__(*handlers, **kwargs)
        self.loop = loop

    def run(self):
        """Start Reactor container and begin processing."""
        self.start()
        self.touch()

    def touch(self):
        """Instruct the reactor container to do processing.

        You might need to call this to startup new sessions. This is already
        handled for create_receiver and create_sender.
        """
        if self.process() and not self.quiesced:
            self.loop.call_soon_threadsafe(self.touch)

    def create_receiver(
            self, context: Union[Connection, Session, Url, str],
            source=None, target=None, name=None, dynamic=False, handler=None,
            options: Union[LinkOption, List[LinkOption]] = None) -> Receiver:
        """Initiate a link to receive messages (subscription).

        Args:
            context: One of: created session, connection with or without
                established session, or url to create session.
            source: Source address.
            target: Target address.
            name: Name of the link.
            dynamic: Wether a dynamic AMQP queue should be generated.
            handler: Custom handler to handle received message.
            options: LinkOptions to further control the attachment.

        Returns:
            Receiver: A Qpid Receiver link over which messages are received.
        """
        receiver = super().create_receiver(context, source, target, name,
                                           dynamic, handler, options)
        self.touch()

        return receiver

    def create_sender(
            self, context: Union[Connection, Session, Url, str],
            target=None, source=None, name=None, handler=None, tags=None,
            options: Union[LinkOption, List[LinkOption]] = None) -> Sender:
        """Initiate a link to send messages.

        Args:
            context: One of: created session, connection with or without
                established session, or url to create session.
            target: Target address.
            source: Source address.
            name: Name of the link.
            handler: Custom handler to handle received message.
            tags:
            options: LinkOptions to further control the attachment.

        Returns:
            Sender: A Qpid Sender link over which messages are sent.
        """
        sender = super().create_sender(context, target, source, name, handler,
                                       tags, options)
        self.touch()

        return sender



import json
from argparse import FileType
from itertools import repeat
from os import EX_DATAERR

from qpid_bow.message import create_message
from qpid_bow.sender import Sender


def message_send_parser(action):
    parser = action.add_parser(
        'send', help="Send one or multiple messages")
    parser.set_defaults(parser=parser)
    parser.set_defaults(func=message_send)

    parser.add_argument('address', metavar='ADDRESS', type=str,
                        help="Queue or exchange address to send a message to")
    parser.add_argument('-b', '--broker-url', type=str, required=False,
                        help="amqp:// or amqps:// URL to the broker")
    parser.add_argument('-r', '--repeat', type=int, required=False, default=1,
                        help="Send same message several times")
    parser.add_argument('-p', '--properties-file', type=FileType('r'),
                        required=False, default=None,
                        help="Message properties file (JSON)")
    parser.add_argument('message', type=str, help="Message to send")


def message_send(args):
    properties = None
    if args.properties_file:
        try:
            properties = json.load(args.properties_file)
        except json.JSONDecodeError:
            print("Invalid JSON properties file\n")
            args.parser.print_help()
            exit(EX_DATAERR)

    message = create_message(args.message.encode(), properties)
    sender = Sender(args.address, args.broker_url)
    sender.queue(repeat(message, args.repeat))
    sender.send()



from os import EX_USAGE

from qpid_bow.management.queue import purge_queue


def queue_purge_parser(action):
    parser = action.add_parser(
        'purge', help="Purge a queue, optionally filtering")
    parser.set_defaults(parser=parser)
    parser.set_defaults(func=queue_purge)

    parser.add_argument('queue', metavar='QUEUE', type=str,
                        help="Queue address")
    parser.add_argument('-b', '--broker-url', type=str, required=False,
                        help="amqp:// or amqps:// URL to the broker")
    parser.add_argument('-l', '--limit', type=int, required=False, default=0,
                        help="Limit the number of messages")
    parser.add_argument('-f', '--filter', type=str, default=None, nargs='?',
                        help="Filter based on property=value")


def queue_purge(args):
    try:
        filter_key, filter_value = args.filter.split('=', 1)
        filter_tuple = (filter_key, filter_value)
    except ValueError:
        print("Invalid filter specified\n")
        args.parser.print_help()
        exit(EX_USAGE)
    except AttributeError:
        filter_tuple = None

    try:
        purge_queue(args.queue, args.limit, filter_tuple, args.broker_url)
    except ValueError as e:
        print(str(e) + '\n')
        args.parser.print_help()
        exit(EX_USAGE)



from unittest import TestCase
from uuid import uuid4

from qpid_bow.config import configure
from qpid_bow.exc import ObjectNotFound, QMF2NotFound, QMF2ObjectExists
from qpid_bow.management.exchange import (
    create_binding,
    create_exchange,
    delete_exchange,
    ExchangeType
)
from qpid_bow.management.queue import (
    create_queue,
    delete_queue,
)

from . import TEST_AMQP_SERVER

CONFIG = {
    'amqp_url': TEST_AMQP_SERVER
}


class TestManagement(TestCase):
    def setUp(self):
        configure(CONFIG)
        self.queue_address = uuid4().hex
        self.exchange_address = uuid4().hex

    def tearDown(self):
        try:
            delete_queue(self.queue_address)
        except QMF2NotFound:
            pass
        try:
            delete_exchange(self.exchange_address)
        except QMF2NotFound:
            pass

    def test_create_queue(self):
        create_queue(self.queue_address, durable=False, auto_delete=True)

    def test_create_queue_already_exists(self):
        create_queue(self.queue_address, durable=False, auto_delete=True)
        with self.assertRaises(QMF2ObjectExists):
            create_queue(self.queue_address, durable=False, auto_delete=True)

    def test_create_exchange(self):
        create_exchange(self.exchange_address,
                        exchange_type=ExchangeType.direct,
                        durable=False)

    def test_create_exchange_already_exists(self):
        create_exchange(self.exchange_address,
                        exchange_type=ExchangeType.direct,
                        durable=False)
        with self.assertRaises(QMF2ObjectExists):
            create_exchange(self.exchange_address,
                            exchange_type=ExchangeType.direct,
                            durable=False)

    def test_create_binding(self):
        create_exchange(self.exchange_address,
                        exchange_type=ExchangeType.topic,
                        durable=False)
        create_queue(self.queue_address, durable=False, auto_delete=True)
        create_binding(self.exchange_address, self.queue_address,
                       binding_name='one_ring')

    def test_create_headers_binding(self):
        create_exchange(self.exchange_address,
                        exchange_type=ExchangeType.headers,
                        durable=False)
        create_queue(self.queue_address, durable=False, auto_delete=True)
        create_binding(self.exchange_address,
                       self.queue_address,
                       headers_match={'type': 'food', 'timing': 'early'},
                       binding_name='breakfast')

    def test_create_binding_match_on_wrong_exchange(self):
        create_exchange(self.exchange_address,
                        exchange_type=ExchangeType.direct,
                        durable=False)
        create_queue(self.queue_address, durable=False, auto_delete=True)
        with self.assertRaises(RuntimeError):
            create_binding(self.exchange_address,
                           self.queue_address,
                           headers_match={'type': 'food', 'timing': 'late'},
                           binding_name='dinner')

    def test_create_binding_no_exchange(self):
        create_queue(self.queue_address, durable=False, auto_delete=True)
        with self.assertRaises(ObjectNotFound):
            create_binding(self.exchange_address, self.queue_address)

    def test_create_binding_no_queue(self):
        create_exchange(self.exchange_address,
                        exchange_type=ExchangeType.direct,
                        durable=False)
        with self.assertRaises(QMF2NotFound):
            create_binding(self.exchange_address, self.queue_address)



"""Exceptions."""


class MessageCorrupt(Exception):
    """Corrupt."""
    pass


class UnroutableMessage(Exception):
    """Origin message has no reply-to address."""
    pass


class RetriableMessage(Exception):
    """Release message back to the queue."""
    pass


class ObjectNotFound(Exception):
    """No object found."""
    def __init__(self, class_name, object_name):
        self.class_name = class_name
        self.object_name = object_name

        super().__init__(f"No {class_name} object found: {object_name}")


class TimeoutReached(Exception):
    """Timeout is reached."""
    pass


class QMF2Exception(Exception):
    """Generic QMF2 exception.

    Args:
        exception_message: Message to identify the reason of exception.
        exception_data: Additional data with error code and error text.
    """
    def __init__(self, exception_message: str, exception_data: dict) -> None:
        super().__init__(exception_message)
        self.exception_data = exception_data

    @staticmethod
    def from_data(exception_data: dict):
        """Try to initialise a specific QMF2Exception based on error code.

        Args:
            exception_data: Additional data with error code and error text.
        """
        error_code = exception_data.get('error_code')
        if error_code == QMF2ObjectExists.error_code:
            if ('object already exists' in
                    exception_data['error_text'].decode()):
                return QMF2ObjectExists(exception_data)
            return QMF2NotFound(exception_data)
        elif error_code == QMF2InvalidParameter.error_code:
            return QMF2InvalidParameter(exception_data)
        elif error_code == QMF2Forbidden.error_code:
            return QMF2Forbidden(exception_data)

        return QMF2Exception("Unknown QMF2 exception", exception_data)


class QMF2NotFound(QMF2Exception):
    """QMF2 object is not found.

    Args:
        exception_data: Additional data with error code and error text.
    """
    error_code = 7

    def __init__(self, exception_data):
        super().__init__("Object not found", exception_data)


class QMF2ObjectExists(QMF2Exception):
    """QMF2 object already exists.

    Args:
        exception_data: Additional data with error code and error text.
    """
    error_code = 7

    def __init__(self, exception_data):
        super().__init__("Object already exists", exception_data)


class QMF2InvalidParameter(QMF2Exception):
    """Invalid parameter is specified.

    Args:
        exception_data: Additional data with error code and error text.
    """
    error_code = 4

    def __init__(self, exception_data):
        super().__init__("Invalid parameter", exception_data)


class QMF2Forbidden(QMF2Exception):
    """Forbidden QMF2 call.

    Args:
        exception_data: Additional data with error code and error text.
    """
    error_code = 6

    def __init__(self, exception_data):
        super().__init__("Forbidden", exception_data)



"""Message utility methods."""

from copy import copy
from typing import Optional, Union

from proton import Message

from qpid_bow import Priority
from qpid_bow.exc import UnroutableMessage


def decode_message(data: bytes) -> Message:
    """Utility method to decode message from bytes.

    Args:
        data: Raw AMQP data in bytes.

    Returns:
        Message: Decoded message.
    """
    message = Message()
    message.decode(data)
    return message


def create_message(body: Union[str, bytes, dict, list],
                   properties: Optional[dict] = None,
                   priority: Priority = Priority.normal) -> Message:
    """Utility method to create message with common attributes.

    Args:
        body: Message body.
        properties: Message properties.
        priority: Message priority.

    Returns:
        Message: Created message.
    """
    message = Message(body=body, durable=True, priority=priority.value,
                      properties=properties or {})
    if isinstance(body, bytes):
        message.content_type = 'application/octet-stream'
    else:
        message.content_type = None

    return message


def create_reply(origin_message: Message,
                 result_data: Union[str, bytes, dict, list]) -> Message:
    """Create reply to origin message with result data.

       Reply messages share the same correlation ID, properties and priority
       with the exception of being marked as reply.

       The address is set to the reply_to address from the origin message
       for usage in a addressless Sender.

    Args:
        origin_message: Origin message we are replying to.
        result_data: Message body of the reply.

    Returns:
        Message: Created reply message.
    """

    if not origin_message.reply_to:
        raise UnroutableMessage("Origin message has no reply-to address")

    reply_message = create_message(result_data,
                                   copy(origin_message.properties),
                                   Priority(origin_message.priority))
    reply_message.address = origin_message.reply_to
    reply_message.correlation_id = origin_message.correlation_id
    reply_message.properties['is_reply'] = True
    return reply_message



from argparse import FileType
from os import (
    EX_DATAERR,
    EX_USAGE,
)

import yaml

from qpid_bow.exc import (
    ObjectNotFound,
    QMF2ObjectExists,
)
from qpid_bow.management.exchange import (
    create_binding,
    create_exchange,
    delete_binding,
    get_binding_keys,
    get_headers_binding_name,
    ExchangeType,
)
from qpid_bow.management.queue import create_queue


def route_config_parser(action):
    parser = action.add_parser(
        'config', help="Configure exchange routing")
    parser.set_defaults(parser=parser)
    parser.set_defaults(func=route_config)

    parser.add_argument('file', metavar='FILE', type=FileType('r'),
                        help="Mapping configuration file (YAML)")
    parser.add_argument('-b', '--broker-url', type=str, required=False,
                        help="amqp:// or amqps:// URL to the broker")
    parser.add_argument('-f', "--force-creation",
                        action='store_true', default=False,
                        help="Force creation of missing exchanges and queues")


def route_config(args):
    try:
        route_mapping = yaml.safe_load(args.file)
    except yaml.YAMLError:
        print("Invalid YAML config mapping\n")
        args.parser.print_help()
        exit(EX_DATAERR)

    for exchange in route_mapping:
        valid_bindings = set()
        if args.force_creation:
            try:
                create_exchange(exchange, ExchangeType.headers,
                                server_url=args.broker_url)
            except QMF2ObjectExists:
                print("Exchange {} already exists, moving on..."
                      .format(exchange))
        for queue in route_mapping[exchange]:
            if args.force_creation:
                try:
                    create_queue(queue, server_url=args.broker_url)
                except QMF2ObjectExists:
                    print("Queue {} already exists, moving on..."
                          .format(queue))
            for binding in route_mapping[exchange][queue]:
                binding_name = get_headers_binding_name(
                    exchange, queue, binding)
                valid_bindings.add((exchange, queue, binding_name))
                # A binding with an existing name will be updated.
                # Since we generate the name from the header match properties,
                # this is equivalent to ignoring.
                try:
                    create_binding(exchange, queue,
                                   headers_match=binding,
                                   binding_name=binding_name,
                                   server_url=args.broker_url)
                except ValueError as e:
                    print(str(e) + '\n')
                    args.parser.print_help()
                    exit(EX_USAGE)
                except ObjectNotFound as e:
                    print(str(e) + '\n')
                    args.parser.print_help()
                    exit(EX_USAGE)

        binding_keys = get_binding_keys(
            exchange, server_url=args.broker_url
        )
        # Delete old bindings
        for binding_key in binding_keys - valid_bindings:
            delete_binding(binding_key[0], binding_key[1], binding_key[2],
                           server_url=args.broker_url)



"""AMQP broker queue management."""
from datetime import timedelta
from typing import (
    Optional,
    Tuple,
)

from qpid_bow.management import (
    create_QMF2_method_invoke,
    get_broker_id,
    get_object,
    handle_QMF2_exception,
)
from qpid_bow.remote_procedure import RemoteProcedure


def reroute_queue(queue_name: str, exchange_name: str,
                  limit: int = 0,
                  message_filter: Optional[Tuple[str, str]] = None,
                  server_url: Optional[str] = None):
    """Reroute messages from a queue to an exchange.

    Args:
        queue_name: Name of queue.
        exchange_name: Name of exchange.
        limit: Limit the amount of messages to reroute.
        message_filter: Filter based on property=value.
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.
    """
    # Gather queue & exchange info + existance check
    queue = get_object('org.apache.qpid.broker', 'queue', queue_name,
                       server_url)
    get_object('org.apache.qpid.broker', 'exchange',
               exchange_name, server_url)

    method_arguments = {'request': limit}  # type: dict

    if exchange_name:
        method_arguments['useAltExchange'] = False
        method_arguments['exchange'] = exchange_name
    else:
        method_arguments['useAltExchange'] = True
        method_arguments['exchange'] = ''

    if message_filter:
        method_arguments['filter'] = _build_message_filter(*message_filter)

    rpc = RemoteProcedure(handle_QMF2_exception,
                          'qmf.default.direct', server_url)
    rpc.call(create_QMF2_method_invoke(queue['_object_id'],
                                       'reroute', method_arguments),
             timedelta(seconds=5))


def purge_queue(queue_name: str,
                limit: int = 0,
                message_filter: Optional[Tuple[str, str]] = None,
                server_url: Optional[str] = None):
    """ Purge a queue on the AMQP broker.

    Args:
        queue_name: Name of queue.
        limit: Limit the amount of messages to purge.
        message_filter: Filter based on property=value.
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.
    """
    queue = get_object('org.apache.qpid.broker', 'queue', queue_name,
                       server_url)
    method_arguments = {'request': limit}  # type: dict
    if message_filter:
        method_arguments['filter'] = _build_message_filter(*message_filter)

    rpc = RemoteProcedure(handle_QMF2_exception,
                          'qmf.default.direct', server_url)
    rpc.call(create_QMF2_method_invoke(queue['_object_id'],
                                       'purge', method_arguments),
             timedelta(seconds=5))


def create_queue(queue_name: str,
                 durable: bool = True,
                 auto_delete: bool = False,
                 priorities: int = 0,
                 extra_properties: Optional[dict] = None,
                 server_url: Optional[str] = None):
    """Create a queue on the AMQP broker.

    Args:
        queue_name: Name of queue.
        durable: Persist the created queue on broker restarts.
        auto_delete: Delete queue after consumer is disconnected from
            broker.
        priorities: The number of priorities to support.
        extra_properties: Additional properties that will be added during
            queue creation.
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.
    """
    method_arguments: dict = {
        'type': 'queue',
        'name': queue_name,
        'properties': {
            'durable': durable,
            'auto-delete': auto_delete,
            'qpid.priorities': priorities
        }
    }

    if extra_properties:
        method_arguments['properties'].update(extra_properties)

    rpc = RemoteProcedure(handle_QMF2_exception,
                          'qmf.default.direct', server_url)
    create_queue_message = create_QMF2_method_invoke(
        get_broker_id(server_url),
        'create', method_arguments)
    rpc.call(create_queue_message, timedelta(seconds=5))


def delete_queue(queue_name: str, server_url: Optional[str] = None):
    """Delete a queue on the AMQP broker.

    Args:
        queue_name: Name of queue.
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.
    """
    rpc = RemoteProcedure(handle_QMF2_exception,
                          'qmf.default.direct', server_url)
    delete_queue_message = create_QMF2_method_invoke(
        get_broker_id(server_url),
        'delete', {
            'type': 'queue',
            'name': queue_name
        }
    )
    rpc.call(delete_queue_message, timedelta(seconds=5))


def _build_message_filter(key: str, value: str) -> dict:
    return {
        'filter_type': 'header_match_str',  # type: ignore
        'filter_params': {'header_key': key, 'header_value': value}
    }



from contextlib import suppress
from datetime import timedelta
from os import environ
from typing import Iterable, Optional
from unittest import TestCase
from uuid import uuid4

from proton import Message

from qpid_bow.config import configure
from qpid_bow.exc import TimeoutReached
from qpid_bow.management.queue import create_queue
from qpid_bow.sender import Sender
from qpid_bow.receiver import Receiver

TEST_AMQP_SERVER = environ.get('AMQP_TEST_SERVERS', 'amqp://127.0.0.1')
CONFIG = {
    'amqp_url': TEST_AMQP_SERVER
}


class MessagingTestBase(TestCase):
    def setUp(self):
        configure(CONFIG)
        self.received_messages = []
        self.expected_messages = []

        queue_address = uuid4().hex
        create_queue(queue_address, durable=False,
                     auto_delete=True, priorities=10,
                     extra_properties={'qpid.auto_delete_timeout': 10})

        self.sender = Sender(queue_address)

        def handle_received_message(message: Message):
            self.received_messages.append(message)
            return True

        self.receiver = Receiver(handle_received_message, queue_address)

    def send_messages(self, messages: Iterable[Message]):
        self.expected_messages.extend(messages)
        self.sender.queue(messages)
        self.sender.send()

    def receive_messages(self):
        with suppress(TimeoutReached):
            self.receiver.receive(timeout=timedelta(seconds=2))

    def check_messages(self, check_identity: bool = True,
                       expected_messages: Optional[Iterable[Message]] = None):
        expected_messages = expected_messages or self.expected_messages
        self.assertEqual(len(self.received_messages), len(expected_messages))
        for received, expected in zip(self.received_messages,
                                      expected_messages):
            self.assertEqual(received.body, expected.body)
            if check_identity:
                # Check identity 100% 
                self.assertEqual(received.id, expected.id)
                self.assertEqual(received.properties, expected.properties)



from unittest import TestCase
from uuid import uuid4

import pytest

from qpid_bow import ReconnectStrategy
from qpid_bow.config import configure
from qpid_bow.message import create_message
from qpid_bow.remote_procedure import RemoteProcedure

from . import TEST_AMQP_SERVER

CONFIG = {
    'amqp_url': TEST_AMQP_SERVER
}


class TestRemoteProcedure(TestCase):
    def setUp(self):
        configure(CONFIG)

    def test_reconnect_strategy_backoff_warning(self):
        with pytest.warns(UserWarning):
            RemoteProcedure(lambda m: True, uuid4().hex,
                            reconnect_strategy=ReconnectStrategy.backoff)

    def test_connection_error(self):
        rpc = RemoteProcedure(lambda m: True, uuid4().hex,
                              server_url='amqp://invalid:invalid@example')
        with self.assertRaises(ConnectionError):
            rpc.call(create_message(b'FOOBAR'))



import re
from os import EX_USAGE

from qpid_bow.management.session import (
    get_outgoing_sessions_by_address,
    get_sessions,
)


def session_outgoing_parser(action):
    parser = action.add_parser(
        'outgoing', help="Dump outgoing sessions")
    parser.set_defaults(parser=parser)
    parser.set_defaults(func=session_outgoing)

    parser.add_argument('-b', '--broker-url', type=str, required=False,
                        help="amqp:// or amqps:// URL to the broker")
    parser.add_argument('-f', '--filter', type=re.compile, required=False,
                        default=None, help="Filter by regex pattern")
    parser.add_argument('-p', '--pretty',
                        required=False, default=False,
                        action='store_true', help="Pretty output")


def session_outgoing(args):
    try:
        sessions = get_sessions(args.broker_url)
    except ValueError as e:
        print(str(e) + '\n')
        args.parser.print_help()
        exit(EX_USAGE)

    outgoing_sessions_by_address = get_outgoing_sessions_by_address(
        args.broker_url)
    for address_name, outgoing_sessions in \
            outgoing_sessions_by_address.items():
        if args.filter and not args.filter.match(address_name):
            continue

        if args.pretty:
            print('╓\n║ Address: {}'.format(address_name))

        num_sessions = len(outgoing_sessions)
        for index, outgoing_session in enumerate(outgoing_sessions):
            last_item = index == num_sessions - 1
            line_char = '┖─' if last_item else '┠─'
            if outgoing_session['session_id'] not in sessions:
                continue
            session = sessions[outgoing_session['session_id']]

            if args.pretty:
                print('║ {}{}'.format(line_char,
                                      session['address']))
                print('║ {}\t Transfers: {}'.format(
                    '' if last_item else '┃',
                    outgoing_session['transfers']))
            else:
                print(session['address'])
        if args.pretty:
            print('╙')



import logging
from argparse import (
    ArgumentParser,
    Namespace,
    _SubParsersAction,
)
from typing import (
    Callable,
    Iterable,
)

from qpid_bow.cli.connection_kill import connection_kill_parser
from qpid_bow.cli.message_receive import message_receive_parser
from qpid_bow.cli.message_send import message_send_parser
from qpid_bow.cli.queue_create import queue_create_parser
from qpid_bow.cli.queue_delete import queue_delete_parser
from qpid_bow.cli.queue_purge import queue_purge_parser
from qpid_bow.cli.queue_reroute import queue_reroute_parser
from qpid_bow.cli.queue_stats import queue_stats_parser
from qpid_bow.cli.route_dump import route_dump_parser
from qpid_bow.cli.route_config import route_config_parser
from qpid_bow.cli.session_outgoing import session_outgoing_parser


def main():
    logging.basicConfig(level=logging.CRITICAL)
    parser = ArgumentParser(prog='qb')
    parser.set_defaults(parser=parser)
    action = parser.add_subparsers(title='commands')

    create_command('queue', 'Manage queues', action,
                   (queue_create_parser, queue_delete_parser,
                    queue_purge_parser, queue_reroute_parser,
                    queue_stats_parser))

    create_command('message', 'Manage messages', action,
                   (message_receive_parser, message_send_parser))

    create_command('route', 'Manage routes', action,
                   (route_dump_parser, route_config_parser))

    create_command('connection', 'Manage connections', action,
                   (connection_kill_parser,))

    create_command('session', 'Manage sessions', action,
                   (session_outgoing_parser,))

    args: Namespace = parser.parse_args()
    if len(args.__dict__) <= 1:
        args.parser.print_help()
    else:
        args.func(args)


def create_command(command: str, helpline: str, action: _SubParsersAction,
                   subparsers: Iterable[Callable]):
    parser = action.add_parser(command, help=helpline)
    parser.set_defaults(parser=parser)
    action = parser.add_subparsers(title='commands')
    for subparser in subparsers:
        subparser(action)


if __name__ == '__main__':
    main()



from unittest import TestCase
from uuid import uuid4

from proton import symbol

from qpid_bow import Priority
from qpid_bow.exc import UnroutableMessage
from qpid_bow.message import create_message, create_reply


class TestMesageCreate(TestCase):
    def test_content_type_none_native_list(self):
        message = create_message([1, 2, 3])
        self.assertEqual(message.content_type, symbol('None'))

    def test_content_type_none_native_dict(self):
        message = create_message({'foo': 'bar'})
        self.assertEqual(message.content_type, symbol('None'))


class TestMesageCreateDefault(TestCase):
    def setUp(self):
        self.properties = {'foo': 'bar'}
        self.message = create_message(b'foobar', self.properties)

    def test_default_priority(self):
        self.assertEqual(Priority(self.message.priority),
                         Priority.normal)

    def test_properties(self):
        self.assertEqual(self.message.properties, self.properties)

    def test_content_type(self):
        self.assertEqual(self.message.content_type, 'application/octet-stream')

    def test_durable(self):
        self.assertTrue(self.message.durable)


class TestMessageCreateReply(TestCase):
    def setUp(self):
        self.properties = {'foo': 'bar'}
        self.origin_message = create_message(
            b'foobar', self.properties, Priority.high)
        self.origin_message.reply_to = 'foobar_address'
        self.origin_message.correlation_id = uuid4()

        self.reply_message = create_reply(self.origin_message, b'foobar-reply')

    def test_unroutable(self):
        with self.assertRaises(UnroutableMessage):
            create_reply(create_message(b'foobar'), b'foobar')

    def test_properties(self):
        expected_properties = {'is_reply': True}
        expected_properties.update(self.origin_message.properties)
        self.assertEqual(self.reply_message.properties, expected_properties)

    def test_content_type(self):
        self.assertEqual(self.reply_message.content_type,
                         'application/octet-stream')

    def test_durable(self):
        self.assertTrue(self.reply_message.durable)

    def test_correlation_id(self):
        self.assertEqual(self.reply_message.correlation_id,
                         self.origin_message.correlation_id)

    def test_address(self):
        self.assertEqual(self.reply_message.address,
                         self.origin_message.reply_to)

    def test_priority(self):
        self.assertEqual(self.reply_message.priority,
                         self.origin_message.priority)



from qpid_bow.receiver import Receiver


def message_receive_parser(action):
    parser = action.add_parser(
        'receive', help="Receive a message")
    parser.set_defaults(parser=parser)
    parser.set_defaults(func=message_receive)

    parser.add_argument('address', metavar='ADDRESS', type=str,
                        help="Queue or exchange address from where to receive "
                             "a message")
    parser.add_argument('-b', '--broker-url', type=str, required=False,
                        help="amqp:// or amqps:// URL to the broker")
    parser.add_argument('-c', '--count', type=int, required=False, default=1,
                        help="How many messages to receive")


def message_receive(args):
    def callback(message):
        print(message.body)
        return True

    receiver = Receiver(callback, args.address, args.broker_url, args.count)
    receiver.receive()



import re

import yaml

from qpid_bow.management import (
    EXCHANGE_ID_PREFIX,
    QUEUE_ID_PREFIX,
)
from qpid_bow.management.exchange import get_exchange_bindings


def route_dump_parser(action):
    parser = action.add_parser(
        'dump', help="Dump exchange route bindings")
    parser.set_defaults(parser=parser)
    parser.set_defaults(func=route_dump)

    parser.add_argument('-b', '--broker-url', type=str, required=False,
                        help="amqp:// or amqps:// URL to the broker")
    parser.add_argument('-f', '--filter', type=re.compile, required=False,
                        default=None, help="Filter by regex pattern")
    parser.add_argument('-y', '--yaml', action='store_true', default=False,
                        help="Output as mapping configuration file")


def route_dump(args):
    try:
        exchange_bindings = get_exchange_bindings(args.broker_url)
    except ValueError as e:
        print(str(e) + '\n')
        args.parser.print_help()
        return

    if args.yaml:
        print(_get_yaml(args, exchange_bindings))
        return

    for exchange_id, bindings in exchange_bindings.items():
        exchange_name = _id_to_name(exchange_id)
        if args.filter and not args.filter.match(exchange_name):
            continue

        print('╓\n║ Exchange: {}'.format(exchange_name))

        num_bindings = len(bindings)
        for index, binding in enumerate(bindings):
            last_item = index == num_bindings - 1
            line_char = '┖─' if last_item else '┠─'
            print('║ {}{}'.format(line_char, _id_to_name(binding['queue_id'])))
            headers_match = binding.get('headers_match')
            if headers_match:
                print('║ {}\t Header match: {}'.format(
                    '' if last_item else '┃',
                    headers_match))
        print('╙')


def _get_yaml(args, exchange_bindings):
    result = {}
    for exchange_id, bindings in exchange_bindings.items():
        exchange_name = _id_to_name(exchange_id)

        if not exchange_name:
            continue

        result[exchange_name] = {}

        if args.filter and not args.filter.match(exchange_name):
            continue

        for binding in bindings:
            queue_name = _id_to_name(binding['queue_id'])
            headers_match: dict = binding.get('headers_match')
            if headers_match:
                if 'x-match' in headers_match:
                    del headers_match['x-match']
                if queue_name not in result[exchange_name]:
                    result[exchange_name][queue_name] = []
                result[exchange_name][queue_name].append(headers_match)

    return yaml.dump(result, default_flow_style=False)


def _id_to_name(value: str) -> str:
    if value.startswith(QUEUE_ID_PREFIX):
        prefix = QUEUE_ID_PREFIX
    else:
        prefix = EXCHANGE_ID_PREFIX
    return value.split(prefix, 1)[1]



"""AMQP broker connection management."""

from datetime import timedelta
from typing import Optional

from proton import Message

from qpid_bow.management import (
    create_QMF2_method_invoke,
    create_QMF2_query,
    handle_QMF2_exception,
)
from qpid_bow.remote_procedure import RemoteProcedure


def get_connection_ids(server_url: Optional[str] = None) -> list:
    """Retrieve connection ids of all established connections to AMQP broker.

    Args:
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.

    Returns:
        List of connections
    """
    connections = []

    def _update_connections(message: Message):
        for item in message.body:
            connections.append(item['_object_id']['_object_name'])
        return True

    rpc = RemoteProcedure(_update_connections, 'qmf.default.direct',
                          server_url)
    binding_query_message = create_QMF2_query('org.apache.qpid.broker',
                                              'connection')
    rpc.call(binding_query_message, timedelta(seconds=5))
    return connections


def kill_connection(connection_id: dict, server_url: Optional[str] = None):
    """Kill connection on AMQP broker.

    Args:
        connection_id: ID of connection.
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.
    """
    rpc = RemoteProcedure(handle_QMF2_exception,
                          'qmf.default.direct', server_url)
    rpc.call(create_QMF2_method_invoke(connection_id, 'close', {}),
             timedelta(seconds=5))



from os import EX_USAGE

from qpid_bow.management.connection import (
    get_connection_ids,
    kill_connection,
)


def connection_kill_parser(action):
    parser = action.add_parser(
        'kill', help="Kills connection")
    parser.set_defaults(parser=parser)
    parser.set_defaults(func=connection_kill)

    parser.add_argument('connections', metavar='CONNECTIONS', type=str,
                        nargs='+', help="Connections to kill, format: IP:PORT")
    parser.add_argument('-b', '--broker-url', type=str, required=False,
                        help="amqp:// or amqps:// URL to the broker")
    parser.add_argument('--commit', action='store_true',
                        default=False, help="Commit to this action")


def connection_kill(args):
    connection_ids = []
    try:
        possible_connection_ids = get_connection_ids(args.broker_url)
    except ValueError as e:
        print(str(e) + '\n')
        args.parser.print_help()
        exit(EX_USAGE)

    for possible_connection_id in possible_connection_ids:
        for requested_connection in args.connections:
            if possible_connection_id.decode().endswith(requested_connection):
                connection_ids.append(possible_connection_id)

    print("{} kill connections:".format('Will' if args.commit else 'Would'))
    for connection_id in connection_ids:
        print(connection_id.decode())
        if args.commit:
            kill_connection({'_object_name': connection_id})



"""AMQP broker exchange management."""

from collections import defaultdict
from copy import copy
from datetime import timedelta
from enum import Enum
from logging import getLogger
from typing import (
    MutableMapping,
    Optional,
    Set,
    Tuple,
)
from uuid import (
    NAMESPACE_URL,
    uuid5,
)

from proton import Message

from qpid_bow.management import (
    create_QMF2_method_invoke,
    create_QMF2_query,
    get_broker_id,
    get_object,
    handle_QMF2_exception,
)
from qpid_bow.remote_procedure import RemoteProcedure

logger = getLogger()
BINDING_NAME_TEMPLATE = 'binding://{exchange}/{queue}/{properties}'


class ExchangeType(Enum):
    """Define type of exchange."""
    direct = "direct"
    topic = "topic"
    fanout = "fanout"
    headers = "headers"


def create_exchange(exchange_name: str,
                    exchange_type: ExchangeType = ExchangeType.direct,
                    durable: bool = True,
                    server_url: Optional[str] = None):
    """Create an exchange on the broker.

    Args:
        exchange_name: Exchange name.
        exchange_type: `direct`, `topic`, `fanout`, `headers`.
        durable: Persist the created exchange on broker restarts.
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.
    """
    rpc = RemoteProcedure(handle_QMF2_exception,
                          'qmf.default.direct', server_url)
    create_exchange_message = create_QMF2_method_invoke(
        get_broker_id(server_url),
        'create', {
            'type': 'exchange',
            'name': exchange_name,
            'properties': {
                'durable': durable,
                'exchange-type': exchange_type.value
            }
        }
    )
    rpc.call(create_exchange_message, timedelta(seconds=5))


def delete_exchange(exchange_name: str, server_url: Optional[str] = None):
    """Delete an exchange on the broker.

    Args:
        exchange_name: Exchange name.
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.
    """
    rpc = RemoteProcedure(handle_QMF2_exception,
                          'qmf.default.direct', server_url)
    delete_exchange_message = create_QMF2_method_invoke(
        get_broker_id(server_url),
        'delete', {
            'type': 'exchange',
            'name': exchange_name
        }
    )
    rpc.call(delete_exchange_message, timedelta(seconds=5))


def create_binding(exchange_name: str, queue_name: str,
                   binding_name=None, headers_match: dict = None,
                   server_url: Optional[str] = None):
    """Create binding between queue and exchange.

    Args:
        exchange_name: Name of exchange.
        queue_name: Name of queue.
        binding_name: Name of binding.
        headers_match: Headers key-value pairs that should be presented on
            message to match the binding. Only for `headers` exchange type.
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.
    """
    exchange = get_object('org.apache.qpid.broker', 'exchange', exchange_name,
                          server_url)
    if headers_match and exchange['_values']['type'].decode() != 'headers':
        raise RuntimeError("Headers match only supported on headers exchange")

    method_arguments: MutableMapping = {
        'type': 'binding',
        'name': '{}/{}'.format(exchange_name, queue_name),
    }

    if binding_name:
        method_arguments['name'] = '{}/{}'.format(method_arguments['name'],
                                                  binding_name)

    if headers_match:
        method_arguments['properties'] = copy(headers_match)
        method_arguments['properties']['x-match'] = 'all'

    rpc = RemoteProcedure(handle_QMF2_exception,
                          'qmf.default.direct', server_url)
    rpc.call(create_QMF2_method_invoke(get_broker_id(server_url),
                                       'create', method_arguments),
             timedelta(seconds=5))


def delete_binding(exchange_name: str, queue_name: str,
                   binding_name: str = None, server_url: Optional[str] = None):
    """Delete a binding on the broker.

    Args:
        exchange_name: Name of exchange.
        queue_name: Name of queue.
        binding_name: Name of binding.
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.
    """
    rpc = RemoteProcedure(handle_QMF2_exception,
                          'qmf.default.direct', server_url)
    method_arguments = {
        'type': 'binding',
        'name': '{}/{}'.format(exchange_name, queue_name)
    }
    if binding_name:
        method_arguments['name'] = '{}/{}'.format(method_arguments['name'],
                                                  binding_name)
    delete_binding_message = create_QMF2_method_invoke(
        get_broker_id(server_url), 'delete', method_arguments
    )
    rpc.call(delete_binding_message, timedelta(seconds=5))


def get_exchange_bindings(server_url: Optional[str] = None) -> dict:
    """Retrieve all exchanges and bindings associated with these exchanges.

    Args:
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.

    Returns:
        dict: A dict mapping between exchange it's bindings.

    Example:
        >>> get_exchange_bindings()
        {'org.apache.qpid.broker:exchange:': [\
{'queue_id': 'org.apache.qpid.broker:queue:examples', 'headers_match': {}}]}
    """
    results: defaultdict = defaultdict(list)

    def _update_bindings(message: Message):
        for item in message.body:
            logger.info("Got binding: %s", item)
            values = item['_values']

            exchange_id = values['exchangeRef']['_object_name'].decode()
            results[exchange_id].append({
                'queue_id': values['queueRef']['_object_name'].decode(),
                'headers_match': values.get('arguments')
            })

        return True

    rpc = RemoteProcedure(_update_bindings, 'qmf.default.direct',
                          server_url)
    binding_query_message = create_QMF2_query('org.apache.qpid.broker',
                                              'binding')
    rpc.call(binding_query_message, timedelta(seconds=5))
    return results


def get_binding_keys(exchange_name: str, queue_name: str = None,
                     server_url: Optional[str] = None
                    ) -> Set[Tuple[str, str, str]]:
    """Retrieve all bindings for specified exchange.

    Args:
        exchange_name: Name of exchange.
        queue_name: Name of queue.
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.

    Returns:
        Set of binding keys.
    """
    result = set()

    def _filter_bindings(message: Message):
        for item in message.body:
            values = item['_values']
            queue_id = values['queueRef']['_object_name'].decode()
            qpid_queue_name = queue_id.rsplit(':', 1)[-1]
            exchange_id = values['exchangeRef']['_object_name'].decode()
            qpid_exchange_name = exchange_id.rsplit(':', 1)[-1]
            if exchange_name == qpid_exchange_name:
                if not queue_name or (queue_name == qpid_queue_name):
                    result.add((
                        qpid_exchange_name,
                        qpid_queue_name,
                        values['bindingKey'].decode()
                    ))
        return True

    rpc = RemoteProcedure(_filter_bindings, 'qmf.default.direct', server_url)
    binding_query_message = create_QMF2_query('org.apache.qpid.broker',
                                              'binding')
    rpc.call(binding_query_message, timedelta(seconds=5))
    return result


def get_headers_binding_name(exchange: str, queue_name: str,
                             headers_match: dict):
    """Generate UUID for exchange, queue and binding.

    Args:
        exchange: Name of exchange.
        queue_name: Name of queue.
        headers_match: Headers key-value pairs that should be presented on
            message to match the binding. Only for `headers` exchange type.

    Returns:
        UUID generated based on specified arguments.
    """
    binding = BINDING_NAME_TEMPLATE.format(
        exchange=exchange,
        queue=queue_name,
        properties='/'.join(
            '{key}/{value}'.format(key=str(key), value=str(value))
            for (key, value) in sorted(headers_match.items())
        )
    )
    return uuid5(NAMESPACE_URL, binding).hex



from os import EX_USAGE

from qpid_bow.management.queue import create_queue
from qpid_bow.exc import (
    QMF2NotFound,
    QMF2ObjectExists,
    TimeoutReached,
)


def queue_create_parser(action):
    parser = action.add_parser(
        'create', help="Create queue")
    parser.set_defaults(parser=parser)
    parser.set_defaults(func=queue_create)

    parser.add_argument('queue', metavar='QUEUE', type=str,
                        help="Queue address")
    parser.add_argument('-b', '--broker-url', type=str, required=False,
                        help="amqp:// or amqps:// URL to the broker")
    parser.add_argument('-d', '--durable',
                        action='store_true', default=False,
                        help="Persist on broker restart")
    parser.add_argument('-p', '--priorities', type=int, required=False,
                        default=0,
                        help="Amount of priorities in queue, default: 0")


def queue_create(args):
    try:
        create_queue(args.queue, args.durable, priorities=args.priorities,
                     server_url=args.broker_url)
    except ValueError as e:
        print(str(e) + '\n')
        args.parser.print_help()
        exit(EX_USAGE)
    except QMF2ObjectExists:
        print(f"Queue '{args.queue}' already exist")
        exit(EX_USAGE)
    except QMF2NotFound:
        print('Object not found, probably not enough permissions to create '
              'queue')
        exit(EX_USAGE)
    except TimeoutReached:
        print('Timeout, server is not available or wrong login/password')
        exit(EX_USAGE)



"""Gather statistics from AMQP broker."""

from datetime import timedelta
from typing import Optional

from proton import Message

from qpid_bow.management import (
    EXCHANGE_ID_PREFIX,
    create_QMF2_query,
)
from qpid_bow.remote_procedure import RemoteProcedure


def queue_statistics(queue_name: Optional[str] = None,
                     include_autodelete: bool = False,
                     server_url: Optional[str] = None) -> dict:
    """Retrieve total messages count and depth for all queues from AMQP broker.

    Args:
        queue_name: Name of queue.
        include_autodelete: Include autodelete queues to output.
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.

    Returns:
        dict: A dict mapping between queue address and dict with total messages
        and queue depth.

    Example:
        >>> queue_statistics(queue_name='examples')
        {'org.apache.qpid.broker:queue:examples': {'name': 'examples', 'total': 96, 'depth': 12}}
    """
    queues = {}

    def _update_queue_stats(message: Message):
        for item in message.body:
            values = item['_values']
            if values['autoDelete'] and not include_autodelete:
                continue  # We are not interested in temp reply queues

            current_queue_name = values['name'].decode()

            if queue_name and current_queue_name != queue_name:
                continue

            queues[item['_object_id']['_object_name'].decode()] = {
                'name': values['name'].decode(),
                'total': int(values['msgTotalEnqueues']),
                'depth': int(values['msgDepth'])
            }
        return True

    rpc = RemoteProcedure(_update_queue_stats, 'qmf.default.direct',
                          server_url)
    queue_query_message = create_QMF2_query('org.apache.qpid.broker',
                                            'queue')
    rpc.call(queue_query_message, timedelta(seconds=15))
    return queues


def exchange_statistics(server_url: Optional[str] = None) -> dict:
    """Retrieve total and dropped amount of messages for exchanges from AMQP broker.

    Args:
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.

    Returns:
        dict: A dict mapping between exchange address and dict with exchange
            name, total messages count and dropped messages count.

    Example:
        >>> exchange_statistics()
        {'org.apache.qpid.broker:exchange:': {'name': '', 'total': ulong(236), 'dropped': ulong(0)}}
    """
    exchanges = {}

    def _update_exchange_stats(message: Message):
        for item in message.body:
            values = item['_values']
            name = values['name'].decode()
            if name.startswith(('qmf', 'qpid', 'amq')):
                continue  # Don't export Qpid/QMF related stats

            exchanges[item['_object_id']['_object_name'].decode()] = {
                'name': name,
                'total': values['msgReceives'],
                'dropped': values['msgDrops']
            }
        return True

    rpc = RemoteProcedure(_update_exchange_stats, 'qmf.default.direct',
                          server_url)
    exchange_query_message = create_QMF2_query('org.apache.qpid.broker',
                                               'exchange')
    rpc.call(exchange_query_message, timedelta(seconds=5))
    return exchanges


def gather_statistics(server_url: Optional[str] = None) -> dict:
    """Retrieve statistics about exchanges and queues from AMQP broker.

    Statistics data includes exchanges and queues. Exchange information
    includes exchange name, total and dropped amount of messages. Queue
    information includes messages count, depth and bindings to exchange.

    Args:
        server_url: Comma-separated list of urls to connect to.
            Multiple can be specified for connection fallback, the first
            should be the primary server.

    Returns:
        dict: Exchange and queue statistics.

    Example:
        >>> gather_statistics()
        {'exchanges': {'org.apache.qpid.broker:exchange:': {'dropped': \
ulong(0), 'name': '', 'total': ulong(251)}}, 'queues': {\
'org.apache.qpid.broker:queue:examples': {'bindings': [{'exchange_id': \
'org.apache.qpid.broker:exchange:', 'name': 'default_route', 'total': 96}], \
'depth': 12, 'name': 'examples', 'total': 96}}}
    """
    stats = {
        'queues': queue_statistics(server_url=server_url),
        'exchanges': exchange_statistics(server_url),
    }

    for queue in stats['queues'].values():
        queue['bindings'] = []

    def _update_binding_stats(message: Message):
        for item in message.body:
            values = item['_values']
            queue_id = values['queueRef']['_object_name'].decode()
            exchange_id = values['exchangeRef']['_object_name'].decode()
            if queue_id not in stats['queues']:
                continue  # Filtered queue
            if exchange_id not in stats['exchanges']:
                continue  # Filtered exchange
            if exchange_id == EXCHANGE_ID_PREFIX:
                continue  # Default exchange stats are broken, reconstruct

            exchange_stats = {
                'name': values['bindingKey'].decode(),
                'exchange_id': exchange_id,
                'total': values['msgMatched']
            }
            stats['queues'][queue_id]['bindings'].append(exchange_stats)
        return True

    rpc = RemoteProcedure(_update_binding_stats, 'qmf.default.direct',
                          server_url)
    binding_query_message = create_QMF2_query('org.apache.qpid.broker',
                                              'binding')
    rpc.call(binding_query_message, timedelta(seconds=15))

    # Reconstruct default route stats
    for queue in stats['queues'].values():
        total_routed = sum((binding['total'] for binding in queue['bindings']))
        queue['bindings'].append({
            'name': 'default_route',
            'exchange_id': EXCHANGE_ID_PREFIX,
            'total': queue['total'] - total_routed
        })

    return stats



import os
import re
import sys

from setuptools import find_packages, setup

if sys.version_info < (3, 6):
    sys.exit('Sorry, Python < 3.6 is not supported')

with open('VERSION') as fh:
    __version__ = fh.read().strip()

build_version = __version__
if 'BUILD' in os.environ:
    spec = re.sub(r'[^0-9]', '', os.environ['BUILD'])
    build_version = "{}-{}".format(__version__, spec)

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst')) as f:
    README = f.read()


requires = [
    'python-qpid-proton>=0.18.1',
    'PyYAML>=3.12'
]

test_requires = requires + [
    'astroid<1.6.0',
    'bandit',
    'mypy==0.501',
    'pylint==1.7.0',
    'pytest',
    'pytest-cov',
]

doc_requires = requires + [
    'sphinx',
    'sphinx-rtd-theme',
    'sphinxcontrib-asyncio',
]

setup(
    name='qpid-bow',
    version=build_version,
    packages=find_packages(),
    install_requires=requires,
    tests_require=test_requires,
    extras_require={
        'test': test_requires,
        'docs': doc_requires,
    },
    include_package_data=True,
    zip_safe=True,
    entry_points={
        'console_scripts': ['qb=qpid_bow.cli.qpid_bow:main']
    },
    description='Set of tools to work with Qpid and a library for '
                'high-level Qpid interaction.',
    long_description=README,
    author='Bynder',
    author_email='info@bynder.com',
    url='https://www.bynder.com/',
    license='MIT',
    keywords='bynder',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development',
        'Topic :: Utilities',
    ],
)



"""Configure qpid-bow."""

from os import environ
from typing import (
    List,
    Mapping,
    Optional,
)

config: dict = {}


def configure(new_config: Mapping):
    """Updates global config with provided mapping.

    Args:
        new_config: Mapping with config data.
    """
    config.update(new_config)


def get_urls(urls: Optional[str] = None) -> List[str]:
    """Retrieves server urls from one of the sources.

    The sources priority comes in the following order: passed arguments,
    global config, AMQP_SERVERS environment variable.

    Args:
        urls: Comma-separated urls.

    Returns:
        List[str]: Returns list of urls to connect to.
    """
    if urls:
        return [url.strip() for url in urls.split(',')]

    if config.get('amqp_url'):
        return [url.strip() for url in config['amqp_url'].split(',')]

    amqp_servers = environ.get('AMQP_SERVERS')

    if amqp_servers:
        environ_urls = []
        for server in amqp_servers.split(','):
            environ_urls.append(server.strip())
        return environ_urls

    raise ValueError('AMQP server url is not configured')



from os import EX_USAGE

from qpid_bow.management.queue import delete_queue
from qpid_bow.exc import (
    QMF2Forbidden,
    QMF2NotFound,
    TimeoutReached,
)


def queue_delete_parser(action):
    parser = action.add_parser(
        'delete', help="Delete queue")
    parser.set_defaults(parser=parser)
    parser.set_defaults(func=queue_delete)

    parser.add_argument('queue', metavar='QUEUE', type=str,
                        help="Queue address")
    parser.add_argument('-b', '--broker-url', type=str, required=False,
                        help="amqp:// or amqps:// URL to the broker")


def queue_delete(args):
    try:
        delete_queue(args.queue, args.broker_url)
    except ValueError as e:
        print(str(e) + '\n')
        args.parser.print_help()
        exit(EX_USAGE)
    except QMF2NotFound:
        print(f"Queue '{args.queue}' not found")
        exit(EX_USAGE)
    except QMF2Forbidden:
        print('Not enough permissions to delete queue')
        exit(EX_USAGE)
    except TimeoutReached:
        print('Timeout, server is not available or wrong login/password')
        exit(EX_USAGE)
